#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>

#include "dat.h"
#include "fns.h"

#include <auth.h>

typedef struct Msg Msg;
struct Msg {
	uchar *buf;
	uchar *end;
	uchar *pos;
	uchar *heap;
};

static Msg *
allocmsg(int n)
{
	Msg *m;

	m = emalloc9p(sizeof(Msg) + n);
	m->pos = m->buf = (uchar *)&m[1];
	m->heap = m->end = m->buf+n;
	return m;
}

static void
resetmsg(Msg *m)
{
	m->pos = m->buf;
	m->heap = m->end;
}


static void
freemsg(Msg *m)
{
	free(m);
}


static char *
strupr(char *s)
{
	char *p;

	for(p = s; *p; p++)
		*p = toupper(*p);
	return s;
}

/* write int */
static void
wint(Msg *m, uint val, int len)
{
	int i;

	assert(m->pos+len < m->heap);
	for(i = 0; i < len; i++){
		*m->pos++ = val;
		val >>= 8;
	}
}

/* read int */
static uint
rint(Msg *m, int len)
{
	int i;
	uint val;

	assert(m->pos+len < m->end);
	val = 0;
	for(i = 0; i < len; i++)
		val |= (*m->pos++ << (8*i));
	return val;
}

/* write mem */
static void
wmem(Msg *m, void *v, int len)
{
	assert(m->pos+len < m->end);
	memcpy(m->pos, v, len);
	m->pos += len;
}

/* read mem */
static void
rmem(Msg *m, void *v, int len)
{
	assert(m->pos+len < m->end);
	memcpy(v, m->pos, len);
	m->pos += len;
}

/* skip reserved data */
static void
skip(Msg *m, int len)
{
	assert(m->pos+len < m->end);
	m->pos += len;
}

/* write utf16 string's params and add its body to the heap */
static void
wHstr(Msg *m, char *str)
{
	Rune r;
	int len;
	char *s;
	ushort *us;

	if(str == nil || *str == 0){
		wint(m, 0, 2);					// length
		wint(m, 0, 2);					// maxlen
		wint(m, 0, 4);					// offset
		return;
	}

	len = utflen(str);
	m->heap -= len * sizeof(ushort);
	assert(m->heap -len > m->pos);
	
	wint(m, len * sizeof(ushort), 2);	// length
	wint(m, len * sizeof(ushort), 2);	// maxlen
	wint(m, m->heap - m->buf, 4);		// offset

	us = (ushort *)m->heap;
	s = str;
	while(*s){
		s += chartorune(&r, s);
		*us++ = r;
	}
}

/* write a data block's params and add it to the heap */
static void
wHmem(Msg *m, void *v, int len)
{
	assert(m->heap -len > m->pos);
	m->heap -= len;

	wint(m, len, 2);					// length
	wint(m, len, 2);					// maxlen
	wint(m, m->heap - m->buf, 4);		// offset

	if(v)
		memcpy(m->heap, v, len);
}


/* read utf16 string's params and get its body to the heap */
static int
rHstr(Msg *m, char *str, int sz)
{
	Rune r;
	ushort *u, *e;
	int i, len, off;

	len = rint(m, 2);
	rint(m, 2);				// maxlen, always the same as len above.
	off = rint(m, 4);

	i = 0;
	u = (ushort *)(m->buf+off);
	e = (ushort *)(m->buf+off+len);
	while(u < e && i < sz){
		r = *u++;
		i += runetochar(str+i, &r);
	}
	str[i] = 0;

	if(i >= sz){
		werrstr("auth_ntlm: name too long\n");
		return -1;
	}

	return 0;
}

static char *
auth_ntlm_phase1(void)
{
	Msg *m;;
	char *p;
	int flags;
	
	/* unicode, need target name, NTLMv1, 56bit encryption */
	flags = (1<<0) | (1<<2) | (1<<9) | (1<<31);

	m = allocmsg(128);
	wmem(m, "NTLMSSP", 8);				// magic
	wint(m, 1, 4);						// message type (1)
	wint(m, flags, 4);					// flags
	wHmem(m, nil, 0);					// windows domain, must be ASCII
	wHmem(m, nil, 0);					// workstation name, must be ASCII
	p = smprint("NTLM %.*[", (int)(m->pos - m->buf), m->buf);
	freemsg(m);
	return p;
}

/*
 * We blindly assume the server will understand utf16 but all 
 * windows boxes do and only windows uses NTLMSSP.
 *
 * Ideally we would key factottum on a realm for the url,
 * Though ntlmssp has somthing it calls a realm this is
 * really an authentication domain, which is in fact, the
 * windows domain name; so we use that.
 *
 * Really we should support NTLMv2 as it is much more difficult
 * to break. However factotum has no client (yet) and
 * the use of NTLMSSP is very limited.
 */
static char *
auth_ntlm_phase2(char *s)
{
	Msg *m;
	int n, flags;
	MSchapreply mcr;
	enum { Maxmsg = 512 };
	uchar nonce[8];
	char *p, domain[64], *host, user[64], magic[8];

	m = allocmsg(Maxmsg);
	if(dec64(m->buf, Maxmsg, s, strlen(s)) == -1){
		werrstr("reply dec64 failed in  NTLM auth reply");
		return nil;
	}

	rmem(m, magic, sizeof(magic));		// magic string
	rint(m, 4);							// message type (2)
	rHstr(m, domain, sizeof(domain));	// domain name
	rint(m, 4);							// flags
	rmem(m, nonce, sizeof(nonce));		// nonce

	n = auth_respond(nonce, sizeof(nonce), user, sizeof(user), &mcr, sizeof mcr,
		auth_getkey, "proto=mschap role=client domain=%s", domain);
	if(n == -1){
		freemsg(m);
		return nil;
	}
	if(! *user){
		werrstr("user= required, but not in factotum entry");
		freemsg(m);
		return nil;
	}

	resetmsg(m);

	/* unicode, need target name, NTLMv1, 56bit encryption */
	flags = (1<<0) | (1<<9) | (1<<31);

	host = strupr(estrdup9p(getenv("sysname")));

	wmem(m, "NTLMSSP", 8);				// magic
	wint(m, 3, 4);						// message type (3)

	wHmem(m, mcr.LMresp, sizeof(mcr.LMresp));
	wHmem(m, mcr.NTresp, sizeof(mcr.NTresp));
	wHstr(m, strupr(domain));
	wHstr(m, user);
	wHstr(m, strupr(host));
	wHstr(m, "");						// no session key
	wint(m, flags, 2);					// flags

	p = smprint("NTLM %.*[", (int)(m->end - m->buf), m->buf);

	free(host);
	freemsg(m);
	return p;
}


static char *
auth_basic(Url *u, char *s)
{
	char *user, *pass, *realm, *x;
	char cred[128], plain[128];
	int n;
	Fmt fmt;
	UserPasswd *up;

	realm = nil;
	user = u->user;
	pass = u->pass;

	if(x = cistrstr(s, "realm="))
		realm = unquote(x+6, &s);
	if(realm == nil)
		return nil;
	up = nil;
	if(user == nil || pass == nil){
		fmtstrinit(&fmt);
		fmtprint(&fmt, " realm=%q", realm);
		if(user)
			fmtprint(&fmt, " user=%q", user);
		if((s = fmtstrflush(&fmt)) == nil)
			return nil;
		up = auth_getuserpasswd(nil, "proto=pass service=http server=%q%s", u->host, s);
		free(s);
		if(up == nil)
			return nil;
		user = up->user;
		pass = up->passwd;
	}
	n = snprint(plain, sizeof(plain), "%s:%s", user ? user : "", pass ? pass : "");
	if(up){
		memset(up->user, 0, strlen(up->user));
		memset(up->passwd, 0, strlen(up->passwd));
		free(up);
	}
	n = enc64(cred, sizeof(cred), (uchar*)plain, n);
	memset(plain, 0, sizeof(plain));
	if(n == -1)
		return nil;
	fmtstrinit(&fmt);
	fmtprint(&fmt, "Basic %s", cred);
	memset(cred, 0, sizeof(cred));
	return fmtstrflush(&fmt);
}

static char *
auth_digest(Url *u, Url *ru, char *method, char *s)
{
	int n;
	char *nonce, *opaque, *user, *x, *realm;
	char chal[1024], ouser[128], resp[2*MD5LEN+1];
	int nchal;
	Fmt fmt;

	user = u->user;
	realm = nonce = opaque = nil;

	if(x = cistrstr(s, "realm="))
		realm = unquote(x+6, &s);
	if(x = cistrstr(s, "nonce="))
		nonce = unquote(x+6, &s);
	if(x = cistrstr(s, "opaque="))
		opaque = unquote(x+7, &s);
	if(realm == nil || nonce == nil)
		return nil;
	fmtstrinit(&fmt);
	fmtprint(&fmt, " realm=%q", realm);
	if(user)
		fmtprint(&fmt, " user=%q", user);
	if((s = fmtstrflush(&fmt)) == nil)
		return nil;
	nchal = snprint(chal, sizeof(chal), "%s %s %U", nonce, method, ru);
	n = auth_respond(chal, nchal, ouser, sizeof ouser, resp, sizeof resp, nil,
		"proto=httpdigest role=client server=%q%s", u->host, s);
	memset(chal, 0, sizeof(chal));
	free(s);
	if(n < 0)
		return nil;
	fmtstrinit(&fmt);
	fmtprint(&fmt, "Digest ");
	fmtprint(&fmt, "username=\"%s\", ", ouser);
	fmtprint(&fmt, "realm=\"%s\", ", realm);
	fmtprint(&fmt, "host=\"%H\", ", u->host);
	fmtprint(&fmt, "uri=\"%U\", ", ru);
	fmtprint(&fmt, "nonce=\"%s\", ", nonce);
	fmtprint(&fmt, "response=\"%s\"", resp);
	if(opaque)
		fmtprint(&fmt, ", opaque=\"%s\"", opaque);
	return fmtstrflush(&fmt);
}

int
authenticate(Url *u, Url *ru, char *method, char *s)
{
	Hauth *a;

	if(!cistrncmp(s, "Basic ", 6)){
		if((s = auth_basic(u, s+6)) == nil)
			return -1;
		u = saneurl(url(".", u));	/* all uris below the requested one */
	}else
	if(!cistrncmp(s, "Digest ", 7)){
		if((s = auth_digest(u, ru, method, s+7)) == nil)
			return -1;
		u = saneurl(url("/", u));	/* BUG: should be the ones in domain= only */
	} else
	if(!cistrcmp(s, "NTLM")){		/* NTLM at the end of the string -> Phase 1 */
		if((s = auth_ntlm_phase1()) == nil)
			return -1;
		u = saneurl(url(".", u));	/* all uris below the requested one */
	}
	else
	if(!cistrncmp(s, "NTLM ", 5)){		/* NTLM with base 64 after -> Phase 2 */
		if((s = auth_ntlm_phase2(s+5)) == nil)
			return -1;
		u = saneurl(url(".", u));	/* all uris below the requested one */
	}
	else
		return -1;

	if(u == nil){
		free(s);
		return -1;
	}
	a = emalloc(sizeof(*a));
	a->url = u;
	a->auth = s;
	qlock(&authlk);
	a->next = hauth;
	hauth = a;
	qunlock(&authlk);

	return 0;
}

